--[[----

	DeclinerPlayerTrack contains all the players recently interacted with.
	Players are cleared as Decliner forgets whispers or they don't spam you after some time.
	Usage, such as in another addon or a WeakAura:

			DeclinerPlayerTrack["action"]["player"]

		is a variable equal to the number of times that action occurred recently from that player.

			#DeclinerPlayerTrack["action"]

		is a variable equal to how many players are currently listed for that action.

	Actions are "whisper", "duel", "pet", guild", "group", "chart", "trade", or "chan".
	"whisper" is the amount of times _you_ whispered that player while the others are the times they were blocked for that action.

	For example, Suprawsom spammed a duel request 5 times, so DeclinerPlayerTrack["duel"]["Suprawsom"] will equal 5, then 4, and so on as time goes by.
	Once the last action is cleared, DeclinerPlayerTrack["duel"]["Suprawsom"] will be nil, allowing Decliner to once again notify you of any further duels from them.
	Another example, you whispered Chanter 9 times due to a conversation, so DeclinerPlayerTrack["whisper"]["Chanter"] will be 9.
	They will be authorized for whatever your options are set to for whispers, such as trading, until that number counts down to nil.

	An example of the second variable, let's say you are standing in front of Orgrimmar on the Area 52 server.
	14 players just spammed you with duel invites at the same time. #DeclinerPlayerTrack["duel"] will equal 14.
	Depending on how many times each player spammed you, that number will count down erratically.
	It may jump directly to nil if all 14 only sent one request or hang around at a lower number for a while when the single requests are cleared.

--------

	DeclinerAllowed is the main player check function, the bouncer, which decides if a player is a stranger or not.
	Usage, such as in another addon or a WeakAura:

			DeclinerAllowed("name")

		Returns true if you recently whispered "name", they're in your group, in your guild, your friend, or on bnet.
		Returns false if they are a complete stranger.

			DeclinerAllowed("name","action")

		Check a player against your settings for that action.
		Actions are "duel", "pet", "guild", "group", "chart", "trade", or "chan".

	For example, DeclinerAllowed("Guildie") will return true because Guildie is in your guild.

	An example of the second one, let's say your options will decline duel requests from guild members and realm friends.
	Mcstabyrouge is a guildie, but not a friend, DeclinerAllowed("Mcstabyrouge","duel") will return false.
	You add them as a friend, DeclinerAllowed("Mcstabyrouge","duel") still returns false.

----]]--

-- addon name and localization table

local a,L=...

-- initialize frames

local DCL=CreateFrame('frame',a)
local DCLPanel=CreateFrame('frame',a..'Panel')
DCL:RegisterEvent('ADDON_LOADED')
DCL:RegisterEvent('PLAYER_LOGIN')
DCL:RegisterEvent('PLAYER_LOGOUT')

-- localize some global functions used

local format=format
local getmetatable=getmetatable
local gsub=gsub
local next=next
local select=select
local setmetatable=setmetatable
local sort=sort
local strlower=strlower
local strmatch=strmatch
local strupper=strupper
local tinsert=tinsert
local tostring=tostring
local tremove=tremove
local type=type

-- initialize functions

local allowed
local blizzardcode
local blizzardoptions
local blizzardoptionslegion
local create
local debug
local Decliner_OnHide
local Decliner_OnShow
local openinvite
local openinvlist
local options
local pattern
local print
local realmstrip
local register
local SetLFGFrame
local setoptions
local SetPetitionFrame
local SetStaticPopup
local SetTradeFrame
local tcopy
local timer

-- initialize tables and variables

local _
local dclconflict={}
local DeclinerDialogs={}
local event={}
local g='|cff00ff00'
local gsnd={}
local initialwarning
local notemp=true
local numbutton=0
local numtext=0
local object={}
local oldSFX
local openinvites
local openinvtrack=0
local player
local r='|cffff0000'
local SPD_={}
local temp={}
local tradeinitiated
local uiinfo={}
local vars={}
local defs={
	enable={chan=true,duel=true,pet=true,guild=true,group=true,chart=true,trade=true,},
	tell={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	group={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	guild={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	friend={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	bnet={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	msg={chan=false,duel=false,pet=false,guild=false,group=false,chart=false,trade=false,},
	chan={emote=true,say=true,yell=true,[1]=true,[2]=true,[26]=true,},
	gen={openinv=false,msgs=false,disable=false,},
	open={'invite','portal',},
	log={[1]=1,},
}
local track={['whisper']={}}
for k in next,defs.enable do
	track[k]={}
end
local file={
	['ch']='Interface\\RaidFrame\\ReadyCheck-Ready',
	['cr']='Interface\\RaidFrame\\ReadyCheck-NotReady',
}

-- global string strip
-- originally from Phanx

function pattern(str)
	str=gsub(str,'|H.-|h.-|h','') -- removes player links
	str=gsub(str,'%%%d?$?s','') -- removes %s
	str=gsub(str,':','') -- removes colons
	str=gsub(str,'%(','') -- removes (
	str=gsub(str,'%)','') -- removes )
	str=gsub(str,'%.','') -- removes periods
	str=gsub(str,'^%s','') -- removes leftover space from beginning of string
	str=gsub(str,'%s$','') -- removes leftover space from end of string
	return str
end

-- Blizzard global strings used, with enUS in comments. parts in quotes are raw strings before pattern.
-- this table is only here to keep strings in one place. it's inefficient to do this, especially since some of these are used once.

local S={
	ADR=ADD..'/'..REMOVE,										--  Add/Remove
	BNF=BATTLENET_OPTIONS_LABEL..' '..FRIENDS,					--  Battle.net Friends
	CHA=CHANNEL,												--  Channel
	CHS=CHANNELS,												--  Channels
	DCA=pattern(ERR_DUEL_CANCELLED),							-- "Duel cancelled."
	DDA=DISABLE..' '..a,										--  Disable Decliner
	DDM=DISABLE..' '..PLAYER_MESSAGES,							--  Disable Player Messages
	DOI=DISABLE..' '..OPENING..' '..GROUP..' ',					--  Disable Opening Group (Invites)
	DRE=pattern(DUEL_REQUESTED),								-- "%s has challenged you to a duel."
	DUE=DUEL,													--  Duel
	EDE=ENABLE..' '..DECLINE,									--  Enable Decline
	EMO=EMOTE,													--  Emote
	GEN=GENERAL,												--  General
	GLD=GUILD,													--  Guild
	GME=GUILD..' '..MEMBERS,									--  Guild Members
	GRE=pattern(ERR_INVITED_ALREADY_IN_GROUP_SS),				-- "|Hplayer:%s|h[%s]|h invited you to a group, but you could not accept because you are already in a group."
	GRI=pattern(ERR_INVITED_TO_GROUP_SS),						-- "|Hplayer:%s|h[%s]|h has invited you to join a group."
	GRO=GROUP,													--  Group
	GRM=GROUP..' '..MEMBERS,									--  Group Members
	GUI=pattern(ERR_INVITED_TO_GUILD_SSS),						-- "|Hplayer:%s|h[%s]|h invites you to join %s."
	INV=pattern(GUILDEVENT_TYPE_INVITE):gsub("^%l",strupper),	-- "%s invites %s"
	LFG=LFG_TITLE,												--  Looking For Group
	PBA=PET_BATTLE_PVP_QUEUE,									--  Pet Battle
	PBD=pattern(ERR_PETBATTLE_DECLINED),						-- "Pet battle invitation declined."
	PBR=pattern(PET_BATTLE_PVP_DUEL_REQUESTED),					-- "%s has challenged you to a pet-battle."
	PLM=PLAYER_MESSAGES,										--  Player Messages
	PTN=pattern(PETITION_TITLE),								-- "Petition: %s"
	REA=strlower(READY),										--  Ready
	REQ=GUILDINFOTAB_APPLICANTS_NONE,							--  Requests
	RFR=pattern(FRIENDS_LIST_REALM)..' '..FRIENDS,				-- "Realm: %s" Friends
	SAY=SAY,													--  Say
	TCA=pattern(ERR_TRADE_CANCELLED),							-- "Trade cancelled."
	TIN=pattern(ERR_INITIATE_TRADE_S),							-- "You have requested to trade with %s."
	TRA=TRADE,													--  Trade
	VER=strlower(GAME_VERSION_LABEL),							--  Version
	WHI=pattern(AUTOCOMPLETE_LABEL_INTERACTED),					-- "(Whispered)"
	YEL=YELL,													--  Yell

	DRM=gsub(DUEL_REQUESTED,'%%%d?$?s',''),						--  gsub duel message down to character name in chat filter
	PBM=gsub(PET_BATTLE_PVP_DUEL_REQUESTED,'%%%d?$?s',''),		--  gsub pet duel message down to character name in chat filter
}

-- internal bouncer, what DeclinerAllowed uses

function allowed(name,option,suppress)
	if not suppress then debug('INTERACTION LOG: '..name..' - '..option) end
	name=realmstrip(name)
	if not suppress then debug('stripped realm: '..name) end
	local bypass=true
	if option then
		if not suppress then debug('    '..name..': checking '..option..' declines') end
		for k in next,defs.enable do
			if option == k then
				bypass=false
				if vars.gen.disable or not vars.enable[option] then
					if not suppress then debug('    '..name..': allowed - '..option..' declines disabled') end
					return true
				end
			end
		end
		if not suppress then debug('    '..name..': '..option..' declines enabled') end
	end
	if bypass or not vars.tell[option] then
		if next(track['whisper']) then
			for k,v in next,track['whisper'] do
				if name == realmstrip(k) and v then
					if not suppress then debug('    '..name..': allowed - whispered') end
					return true
				end
			end
			if not suppress then debug('    '..name..': did not recently whisper') end
		end
		if not suppress then debug('    '..name..': no one recently whispered') end
	else if not suppress then debug('    '..name..': tell disabled for '..option) end
	end
	if bypass or (option ~= 'group' and not vars.group[option]) then
		if IsInRaid() then
			for i=1,40 do
				if (UnitName('raid'..i)) then
					if name == realmstrip((UnitName('raid'..i))) then
						if not suppress then debug('    '..name..': allowed - raid member') end
						return true
					end
				end
			end
			if not suppress then debug('    '..name..': not a raid member') end
		elseif IsInGroup() then
			for i=1,4 do
				if (UnitName('party'..i)) then
					if name == realmstrip((UnitName('party'..i))) then
						if not suppress then debug('    '..name..': allowed - party member') end
						return true
					end
				end
			end
			if not suppress then debug('    '..name..': not a party member') end
		else if not suppress then debug('    '..name..': not in a group') end
		end
	else if option ~= 'group' and not suppress then debug('    '..name..': group disabled for '..option) end
	end
	if bypass or (option ~= 'guild' and not vars.guild[option]) then
		if IsInGuild() then
			for i=1,(GetNumGuildMembers()) do
				if name == realmstrip((GetGuildRosterInfo(i))) then
					if not suppress then debug('    '..name..': allowed - guild member') end
					return true
				end
			end
			if not suppress then debug('    '..name..': not a guild member') end
		else if not suppress then debug('    '..name..': not in a guild') end
		end
	else if option ~= 'guild' and not suppress then debug('    '..name..': guild disabled for '..option) end
	end
	if bypass or not vars.friend[option] then
		for i=1,(GetNumFriends()) do
			if name == realmstrip((GetFriendInfo(i))) then
				if not suppress then debug('    '..name..': allowed - realm friend') end
				return true
			end
		end
		if not suppress then debug('    '..name..': not a realm friend') end
	else if not suppress then debug('    '..name..': friend disabled for '..option) end
	end
	if bypass or not vars.bnet[option] then
		for i=1,(BNGetNumFriends()) do
			for j=1,(BNGetNumFriendGameAccounts(i)) do
				local _,charname,client=BNGetFriendGameAccountInfo(i,j)
				if client == 'WoW' then
					if name == realmstrip(charname) then
						if not suppress then debug('    '..name..': allowed - battle.net friend') end
						return true
					end
				end
			end
		end
		if not suppress then debug('    '..name..': not a bnet friend') end
	else if not suppress then debug('    '..name..': bnet disabled for '..option) end
	end
	if not suppress then debug('    '..name..': declined') end
	return false
end

-- debug handler

function debug(message,...)
	local args={n=select('#',...),...}
	local argstr=''
	for i=1,args.n do
		if argstr == '' then argstr=' - args: '..tostring(args[i])
		else argstr=argstr..', '..tostring(args[i]) end
	end
	tinsert(vars.log,message..argstr)
	if #vars.log > 99 then tremove(vars.log,2) end
	vars.log[1]=#vars.log
	if Decliner_SV then Decliner_SV.log=tcopy(vars.log) end
end

-- allows all group invites after player last said certain words
-- changing zones or joining a group will cancel this effect
-- effect ends after 5 minutes since last trigger
-- being in a group will not trigger this effect

function openinvite(message,sender)
	if vars.gen.openinv or IsInGroup() or sender ~= player then return end
	for _,v in next,vars.open do
		if strmatch(message,'^'..v..'$')
		or strmatch(message,'^'..v..'[^%w].*')
		or strmatch(message,'.*[^%w]'..v..'$')
		or strmatch(message,'^.*[^%w]'..v..'[^%w].*$') then
			openinvites=true
			timer('openinv')
		end
	end
end

-- chat output
-- the idea for faking a yellow system message came from Jaliborc's Scrap
-- but they hook into the MessageHandler function while I use AddMessage

function print(message,reason,name,other)
	if vars.gen.disable or vars.gen.msgs then return end
	if reason then
		local short=realmstrip(name)
		if vars.msg[reason] or timer('player',reason,name) > 1 then
		debug('    '..short..': spammed '..track[reason][short]..' times recently') return end
		message=format(message,"|Hplayer:"..name.."|h["..name.."]|h",other or nil)
		debug('    '..short..': first recent occurrence')
	end
	local notfound,info=true,ChatTypeInfo['SYSTEM']
	for i=1,NUM_CHAT_WINDOWS do
		if _G['ChatFrame'..i]:IsEventRegistered('CHAT_MSG_SYSTEM') then
			notfound=false
			_G['ChatFrame'..i]:AddMessage(r..a..': |r'..message,info.r,info.g,info.b,info.id)
		end
	end
	if notfound then DEFAULT_CHAT_FRAME:AddMessage(r..a..': |r'..message,info.r,info.g,info.b,info.id) end
end

-- strips realms from character names

function realmstrip(str)
	str=gsub(str,'%-.*','')
	return str
end

-- following function from http://lua-users.org/wiki/CopyTable
-- this copies the values of a table into another table instead of creating a reference

function tcopy(src)
	local copy
	if type(src) == 'table' then
		copy={}
		for k,v in next,src do
			copy[tcopy(k)]=tcopy(v)
		end
		setmetatable(copy,tcopy(getmetatable(src)))
	else copy=src end
	return copy
end

-- timer to control open invites and list of players for spamming and whispers

function timer(which,reason,name)
	if which == 'global' then
		if reason and (not DeclinerAllowed or not DeclinerPlayerTrack or not Decliner_SV) then
		debug('conflict with unknown addon') end
		DeclinerAllowed=function(n,o,s)return allowed(n,o,s)end
		DeclinerPlayerTrack=tcopy(track)
		C_Timer.After(1,function()timer('global',true)end)
	end
	if which == 'openinv' then
		openinvtrack=openinvtrack+1
		C_Timer.After(300,function()
			if openinvtrack == 1 then openinvites=false end
			openinvtrack=openinvtrack-1
		end)
	end
	if which == 'player' then
		name=realmstrip(name)
		if not track[reason][name] then
			track[reason][name]=1
			timer('playerloop',reason,name)
		else
			track[reason][name]=track[reason][name]+1
		end
		return track[reason][name]
	end
	if which == 'playerloop' then
		C_Timer.After(60,function()
			if track[reason][name] == 1 then
				track[reason][name]=nil
			else
				track[reason][name]=track[reason][name]-1
				timer('playerloop',reason,name)
			end
		end)
	end
end

-- event blocks

function event:ADDON_LOADED(addon)
	if addon == a then
		if not Decliner_SV then vars=tcopy(defs) else vars=tcopy(Decliner_SV) end
		for k1,v1 in next,defs do for k2,v2 in next,v1 do
			if vars[k1] == nil then vars[k1]=defs[k1] end
			if vars[k1][k2] == nil then vars[k1][k2]=defs[k1][k2] end
			if type(vars[k1][k2]) ~= type(v2) then vars[k1][k2]=defs[k1][k2] end
		end end
		for k1,v1 in next,vars do
			if defs[k1] == nil then vars[k1]=nil end
		end
		for k1,v1 in next,vars do for k2,v2 in next,v1 do
			if defs[k1][k2] == nil and (k1 ~= 'log' and k1 ~= 'open') then vars[k1][k2]=nil end
		end end
		Decliner_SV=nil
		Decliner_SV=tcopy(vars)
		options()
		register()
		blizzardoptions()
		timer('global')
	else
		if not DeclinerAllowed or not DeclinerPlayerTrack or not Decliner_SV then
			tinsert(dclconflict,addon)
			if initialwarning then debug("conflict with "..dclconflict[#dclconflict]) end
		end
	end
end

function event:PLAYER_LOGIN()
	oldSFX=GetCVar('Sound_EnableSFX')
	local name,realm=UnitFullName('player')
	player=name..'-'..realm
	GuildRoster()
	ShowFriends()
	RegisterAddonMessagePrefix('DCLDBG')
	SPD_['CHAT_CHANNEL_INVITE']=tcopy(StaticPopupDialogs['CHAT_CHANNEL_INVITE'])
	SPD_['DUEL_REQUESTED']=tcopy(StaticPopupDialogs['DUEL_REQUESTED'])
	SPD_['PARTY_INVITE']=tcopy(StaticPopupDialogs['PARTY_INVITE'])
	SPD_['PARTY_INVITE_XREALM']=tcopy(StaticPopupDialogs['PARTY_INVITE_XREALM'])
	SPD_['PET_BATTLE_PVP_DUEL_REQUESTED']=tcopy(StaticPopupDialogs['PET_BATTLE_PVP_DUEL_REQUESTED'])
	print(S['VER']..' '..GetAddOnMetadata(a,'Version')..' '..S['REA'])
	for _,v in next,dclconflict do debug("conflict with "..v) end
	initialwarning=true
end

function event:PLAYER_LOGOUT()
	Decliner_SV=nil
	Decliner_SV=tcopy(vars)
end

function event:CHANNEL_INVITE_REQUEST(...)
	local _,frame=StaticPopup_Visible('CHAT_CHANNEL_INVITE')
	if frame then HideUIPanel(frame) end
	local chan,name=...
	if allowed(name,'chan') then
		if frame then
			SetStaticPopup(frame,'default')
			ShowUIPanel(frame)
		end
	else
		DeclineChannelInvite(chan)
		print(L['DECLINED_CHAN'],'chan',name,chan)
	end
	debug('    CHAN',...)
end

function event:CHAT_MSG_ADDON(prefix,message,channel,sender)
	if not (prefix == 'DCLDBG' and message == a and channel == 'WHISPER') then return end
	SendAddonMessage('DCLDBG',vars.log[1],'WHISPER',sender)
	local index=2
	C_Timer.NewTicker(0.1,function()
		SendAddonMessage('DCLDBG',vars.log[index],'WHISPER',sender)
		index=index+1
	end,#vars.log-1)
end

function event:CHAT_MSG_CHANNEL(message,sender,_,_,_,_,ID)
	if vars.chan[ID] then openinvite(message,sender) end
end

function event:CHAT_MSG_EMOTE(message,sender)
	if vars.chan.emote then openinvite(message,sender) end
end

function event:CHAT_MSG_SAY(message,sender)
	if vars.chan.say then openinvite(message,sender) end
end

function event:CHAT_MSG_SYSTEM(message)
	if strmatch(message,S['TIN']) then
		tradeinitiated=true
	end
end

function event:CHAT_MSG_WHISPER_INFORM(_,target)
	timer('player','whisper',target)
end

function event:CHAT_MSG_YELL(message,sender)
	if vars.chan.yell then openinvite(message,sender) end
end

function event:CVAR_UPDATE(cvar,value)
	oldSFX=GetCVar('Sound_EnableSFX')
	if cvar == 'BLOCK_TRADES' then
		if value == '0' and not vars.enable.trade then
			vars.enable.trade=true
			register()
			setoptions()
			blizzardoptions()
			print(L['X_BLOCK_TRADES_OFF'])
		end
		if value == '1' and vars.enable.trade then
			vars.enable.trade=false
			register()
			setoptions()
			blizzardoptions()
			print(L['X_BLOCK_TRADES_ON'])
		end
	end
	if cvar == 'BLOCK_CHAT_CHANNEL_INVITE' then
		if value == '0' and not vars.enable.chan then
			vars.enable.chan=true
			register()
			setoptions()
			blizzardoptions()
			print(L['X_BLOCK_CHANNELS_OFF'])
		end
		if value == '1' and vars.enable.chan then
			vars.enable.chan=false
			register()
			setoptions()
			blizzardoptions()
			print(L['X_BLOCK_CHANNELS_ON'])
		end
	end
end

function event:DISABLE_DECLINE_GUILD_INVITE()
	if vars.enable.guild then return end
	vars.enable.guild=true
	register()
	setoptions()
	blizzardoptions()
	print(L['X_BLOCK_GUILDS_OFF'])
end

function event:DUEL_REQUESTED(...)
	local _,frame=StaticPopup_Visible('DUEL_REQUESTED')
	if frame then HideUIPanel(frame) end
	local name=...
	if allowed(name,'duel') then
		if frame then
			SetStaticPopup(frame,'default')
			ShowUIPanel(frame)
			PlaySoundKitID(880)
		end
		uiinfo['d']=false
	else
		uiinfo['d']=true
		CancelDuel()
		print(L['DECLINED_DUEL'],'duel',name)
	end
	debug('    DUEL',...)
end

function event:ENABLE_DECLINE_GUILD_INVITE()
	if not vars.enable.guild then return end
	vars.enable.guild=false
	register()
	setoptions()
	blizzardoptions()
	print(L['X_BLOCK_GUILDS_ON'])
end

function event:GROUP_ROSTER_UPDATE()
	openinvites=false
end

function event:GUILD_INVITE_REQUEST(...)
	SetCVar('Sound_EnableSFX','0')
	local name,guild=...
	if allowed(name,'guild') then
		C_Timer.After(0.5,function()
			SetCVar('Sound_EnableSFX',oldSFX)
			PlaySoundKitID(888)
		end)
	else
		StaticPopupSpecial_Hide(GuildInviteFrame)
		DeclineGuild()
		print(L['DECLINED_GUILD'],'guild',name,guild)
		C_Timer.After(0.5,function()SetCVar('Sound_EnableSFX',oldSFX)end)
	end
	debug('    GUILD',...)
end

function event:PARTY_INVITE_REQUEST(...)
	local name,lfg1,lfg2,lfg3,xrlm=...
	local which=xrlm and 'PARTY_INVITE_XREALM' or 'PARTY_INVITE'
	local _,frame=StaticPopup_Visible(which)
	if frame then HideUIPanel(frame) end
	if lfg1 or lfg2 or lfg3 then LFGInvitePopup:Hide() end
	if openinvites or allowed(name,'group') then
		if frame then
			SetStaticPopup(frame,'default')
			ShowUIPanel(frame)
			PlaySoundKitID(880)
			StaticPopupDialogs[which].OnHide=function(self)
				if not self.inviteAccepted then
					DeclineGroup()
					self:Hide()
				end
				StaticPopupDialogs[which]=tcopy(DeclinerDialogs[which])
			end
		elseif lfg1 or lfg2 or lfg3 then
			SetLFGFrame('default')
			LFGInvitePopup:Show()
		end
	else
		DeclineGroup()
		print(L['DECLINED_GROUP'],'group',name)
	end
	debug('    GROUP',...)
end

function event:PET_BATTLE_PVP_DUEL_REQUESTED(...)
	local _,frame=StaticPopup_Visible('PET_BATTLE_PVP_DUEL_REQUESTED')
	if frame then HideUIPanel(frame) end
	local name=...
	if allowed(name,'pet') then
		if frame then
			SetStaticPopup(frame,'default')
			ShowUIPanel(frame)
			PlaySoundKitID(880)
		end
		uiinfo['p']=false
	else
		uiinfo['p']=true
		C_PetBattles.CancelPVPDuel()
		print(L['DECLINED_PET'],'pet',name)
	end
	debug('    PET',...)
end

function event:PETITION_SHOW(...)
	HideUIPanel(PetitionFrame)
	local ptype,_,_,_,name,originator=GetPetitionInfo()
	if originator or allowed(name,'chart') then
		SetPetitionFrame('default')
		ShowUIPanel(PetitionFrame)
	else
		ClosePetition()
		if ptype == 'guild' then print(L['DECLINED_CHART_GUILD'],'chart',name)
		else print(L['DECLINED_CHART_NIL'],'chart',name) end
	end
	debug('    CHART',...)
end

function event:TRADE_CLOSED()
	tradeinitiated=false
end

function event:TRADE_SHOW(...)
	HideUIPanel(TradeFrame)
	-- thank you to http://wowprogramming.com/docs/api_types#unitID for that vague Trade UI note next to 'npc'
	local name,realm=UnitName('npc')
	if realm then name=name..'-'..realm end
	if tradeinitiated or allowed(name,'trade') then
		SetTradeFrame('default')
		ShowUIPanel(TradeFrame)
		uiinfo['t']=false
	else
		uiinfo['t']=true
		CloseTrade()
		print(L['DECLINED_TRADE'],'trade',name)
	end
	debug('    TRADE',name,...)
end

function event:UI_INFO_MESSAGE(arg1,arg2)
	local msg=arg2 or arg1 -- WoD uses arg1 for the message with arg2 nil, Legion uses arg2 with arg1 an unknown number
	if strmatch(msg,S['DCA']) and uiinfo['d'] then UIErrorsFrame:Clear() end
	if strmatch(msg,S['PBD']) and uiinfo['p'] then UIErrorsFrame:Clear() end
	if strmatch(msg,S['TCA']) and uiinfo['t'] then UIErrorsFrame:Clear() end
end

function event:ZONE_CHANGED_NEW_AREA()
	openinvites=false
end

-- calls an event function when that event triggers

DCL:SetScript('OnEvent',function(s,e,...)event[e](s,...)end)

-- chat filter that blocks associated system messages if that action is declined

local DECLINER_CHAT_FILTER=function(_,_,msg)
	if vars.gen.disable then return end
	if strmatch(msg,S['DRE']) and vars.enable.duel then
		local name=gsub(msg,S['DRM'],'')
		if not allowed(name,'duel',true) then return true end
	end
	if strmatch(msg,S['PBR']) and vars.enable.pet then
		local name=gsub(msg,S['PBM'],'')
		if not allowed(name,'pet',true) then return true end
	end
	if strmatch(msg,S['GUI']) and vars.enable.guild then
		local name=strmatch(msg,'|h%[(.-)%]|h')
		if not allowed(name,'guild',true) then return true end
	end
	if strmatch(msg,S['GRI']) and vars.enable.group then
		local name=strmatch(msg,'|h%[(.-)%]|h')
		if not allowed(name,'group',true) then return true end
	end
	if strmatch(msg,S['GRE']) and vars.enable.group then
		local name=strmatch(msg,'|h%[(.-)%]|h')
		if not allowed(name,'group',true) then
		print(L['DECLINED_GROUP'],'group',name) return true end
	end
end
ChatFrame_AddMessageEventFilter('CHAT_MSG_SYSTEM',DECLINER_CHAT_FILTER)

-- Blizzard code replacements (these are injected into the UI, so direct usage of global strings the UI uses instead of S table)
-- all of these are basically alterations of Blizzard's Lua or changing Blizzard's frame scripts
-- you can view Blizzard's Lua with these instructions: http://wow.gamepedia.com/Viewing_Blizzard's_interface_code

DeclinerDialogs['CHAT_CHANNEL_INVITE']={
	text=CHAT_INVITE_NOTICE_POPUP,
	button1=ACCEPT,
	button2=DECLINE,
	OnShow=function(self)
		StaticPopupDialogs['CHAT_CHANNEL_INVITE'].inviteAccepted=nil
	end,
	OnAccept=function(self,data)
		local name=data
		local zoneChannel,channelName=JoinPermanentChannel(name,nil,DEFAULT_CHAT_FRAME:GetID(),1)
		if (channelName) then
			name=channelName
		end
		if (not zoneChannel) then
			return
		end
		local i=1
		while (DEFAULT_CHAT_FRAME.channelList[i]) do
			i=i+1
		end
		DEFAULT_CHAT_FRAME.channelList[i]=name
		DEFAULT_CHAT_FRAME.zoneChannelList[i]=zoneChannel
	end,
	EditBoxOnEnterPressed=function(self,data)
		local name=data
		local zoneChannel,channelName=JoinPermanentChannel(name,nil,DEFAULT_CHAT_FRAME:GetID(),1)
		if (channelName) then
			name=channelName
		end
		if (not zoneChannel) then
			return
		end
		local i=1
		while (DEFAULT_CHAT_FRAME.channelList[i]) do
			i=i+1
		end
		DEFAULT_CHAT_FRAME.channelList[i]=name
		DEFAULT_CHAT_FRAME.zoneChannelList[i]=zoneChannel
		StaticPopupDialogs['CHAT_CHANNEL_INVITE'].inviteAccepted=1
		self:GetParent():Hide()
	end,
	EditBoxOnEscapePressed=function(self,data)
		self:GetParent():Hide()
	end,
	OnCancel=function(self,data)
		local chanName=data
		DeclineChannelInvite(chanName)
	end,
	timeout=CHANNEL_INVITE_TIMEOUT,
	whileDead=1,
	hideOnEscape=1,
}

DeclinerDialogs['DUEL_REQUESTED']={
	text=DUEL_REQUESTED,
	button1=ACCEPT,
	button2=DECLINE,
	OnAccept=function(self)
		AcceptDuel()
	end,
	OnCancel=function(self)
		CancelDuel()
	end,
	timeout=STATICPOPUP_TIMEOUT,
	hideOnEscape=1
}

DeclinerDialogs['PET_BATTLE_PVP_DUEL_REQUESTED']={
	text=PET_BATTLE_PVP_DUEL_REQUESTED,
	button1=ACCEPT,
	button2=DECLINE,
	OnAccept=function(self)
		C_PetBattles.AcceptPVPDuel()
	end,
	OnCancel=function(self)
		C_PetBattles.CancelPVPDuel()
	end,
	timeout=STATICPOPUP_TIMEOUT,
	hideOnEscape=1
}

DeclinerDialogs['PARTY_INVITE']={
	text=INVITATION,
	button1=ACCEPT,
	button2=DECLINE,
	OnShow=function(self)
		self.inviteAccepted=nil
	end,
	OnAccept=function(self)
		AcceptGroup()
		self.inviteAccepted=1
	end,
	OnCancel=function(self)
		DeclineGroup()
	end,
	OnHide=function(self)
		if (not self.inviteAccepted) then
			self:Hide()
		end
	end,
	timeout=STATICPOPUP_TIMEOUT,
	whileDead=1,
	hideOnEscape=1
}

DeclinerDialogs['PARTY_INVITE_XREALM']={
	text=INVITATION_XREALM,
	button1=ACCEPT,
	button2=DECLINE,
	OnShow=function(self)
		self.inviteAccepted=nil
	end,
	OnAccept=function(self)
		AcceptGroup()
		self.inviteAccepted=1
	end,
	OnCancel=function(self)
		DeclineGroup()
	end,
	OnHide=function(self)
		if (not self.inviteAccepted) then
			self:Hide()
		end
	end,
	timeout=STATICPOPUP_TIMEOUT,
	whileDead=1,
	hideOnEscape=1
}

function Decliner_OnHide(frame)
	local dialog=frame.which
	if dialog and
	((dialog == 'DUEL_REQUESTED' and vars.enable.duel) or
	(dialog == 'PET_BATTLE_PVP_DUEL_REQUESTED' and vars.enable.pet) or
	((dialog == 'PARTY_INVITE' or dialog == 'PARTY_INVITE_XREALM') and vars.enable.group) or
	(dialog == 'CHAT_CHANNEL_INVITE' and vars.enable.chan)) then
		StaticPopup_CollapseTable()
		local OnHide=dialog.OnHide
		if (OnHide) then
			OnHide(frame,frame.data)
		end
		if (dialog.enterClicksFirstButton) then
			frame:SetScript('OnKeyDown',nil)
		end
	else
		StaticPopup_OnHide(frame)
	end
end

function Decliner_OnShow(frame)
	local dialog=frame.which
	if dialog and
	((dialog == 'DUEL_REQUESTED' and vars.enable.duel) or
	(dialog == 'PET_BATTLE_PVP_DUEL_REQUESTED' and vars.enable.pet) or
	((dialog == 'PARTY_INVITE' or dialog == 'PARTY_INVITE_XREALM') and vars.enable.group) or
	(dialog == 'CHAT_CHANNEL_INVITE' and vars.enable.chan)) then
		local OnShow=dialog.OnShow
		if (OnShow) then
			OnShow(frame,frame.data)
		end
		if (dialog.enterClicksFirstButton) then
			frame:SetScript('OnKeyDown',StaticPopup_OnKeyDown)
		end
	else
		StaticPopup_OnShow(frame)
	end
end

function SetLFGFrame(action)
	LFGInvitePopup:SetScript('OnShow',nil)
	LFGInvitePopup:SetScript('OnHide',nil)
	if action == 'default' and vars.gen.disable then
		LFGInvitePopup:SetScript('OnShow',function()
			PlaySoundKitID(880)
		end)
		LFGInvitePopup:SetScript('OnHide',function()
			PlaySoundKitID(851)
		end)
	elseif action == 'default' then
		LFGInvitePopup:SetScript('OnShow',function()
			PlaySoundKitID(880)
		end)
		LFGInvitePopup:SetScript('OnHide',function()
			PlaySoundKitID(851)
			if vars.enable.group then SetLFGFrame('decline') end
		end)
	end
end

function SetPetitionFrame(action)
	PetitionFrame:SetScript('OnShow',nil)
	PetitionFrame:SetScript('OnHide',nil)
	if action == 'decline' then
		PetitionFrame:SetScript('OnHide',function()
			StaticPopup_Hide('CONFIRM_GUILD_CHARTER_SIGNATURE')
		end)
	elseif action == 'default' and vars.gen.disable then
		PetitionFrame:SetScript('OnShow',function()
			PlaySoundKitID(875)
		end)
		PetitionFrame:SetScript('OnHide',function()
			PlaySoundKitID(876)
			StaticPopup_Hide('CONFIRM_GUILD_CHARTER_SIGNATURE')
			ClosePetition()
		end)
	elseif action == 'default' then
		PetitionFrame:SetScript('OnShow',function()
			PlaySoundKitID(875)
		end)
		PetitionFrame:SetScript('OnHide',function()
			PlaySoundKitID(876)
			StaticPopup_Hide('CONFIRM_GUILD_CHARTER_SIGNATURE')
			ClosePetition()
			if vars.enable.chart then SetPetitionFrame('decline') end
		end)
	end
end

function SetStaticPopup(frame,action)
	frame:SetScript('OnShow',nil)
	frame:SetScript('OnHide',nil)
	if action == 'decline' then
		frame:SetScript('OnShow',function()
			Decliner_OnShow(frame)
		end)
		frame:SetScript('OnHide',function()
			Decliner_OnHide(frame)
		end)
	elseif action == 'default' and vars.gen.disable then
		frame:SetScript('OnShow',function()
			StaticPopup_OnShow(frame)
		end)
		frame:SetScript('OnHide',function()
			StaticPopup_OnHide(frame)
		end)
	elseif action == 'default' then
		frame:SetScript('OnShow',function()
			StaticPopup_OnShow(frame)
		end)
		frame:SetScript('OnHide',function()
			StaticPopup_OnHide(frame)
			if vars.enable.duel or vars.enable.pet or vars.enable.group or vars.enable.chan then
				SetStaticPopup(frame,'decline')
			end
		end)
	end
end

function SetTradeFrame(action)
	TradeFrame:SetScript('OnShow',nil)
	TradeFrame:SetScript('OnHide',nil)
	if action == 'default' and vars.gen.disable then
		TradeFrame:SetScript('OnShow',function()
			TradeFrame_OnShow(self)
			PlaySoundKitID(839)
		end)
		TradeFrame:SetScript('OnHide',function()
			TradeFrame_OnHide()
			PlaySoundKitID(840)
		end)
	elseif action == 'default' then
		TradeFrame:SetScript('OnShow',function()
			TradeFrame_OnShow(self)
			PlaySoundKitID(839)
		end)
		TradeFrame:SetScript('OnHide',function()
			TradeFrame_OnHide()
			PlaySoundKitID(840)
			if vars.enable.trade then SetTradeFrame('decline') end
		end)
	end
end

-- disables the block trades, block guild invites, and block channel invites options and edits their display text based on user options
-- Decliner won't be able to let authorized players through if these options are enabled

function blizzardoptions()
	if tostring(format('%4$s',GetBuildInfo())) ~= '60200' then blizzardoptionslegion() else
		if vars.enable.trade and not vars.gen.disable then
			_G['InterfaceOptionsControlsPanelBlockTradesText']:SetFontObject('GameFontDisableLeft')
			_G['InterfaceOptionsControlsPanelBlockTradesText']:SetText(BLOCK_TRADES..' ('..a..')')
			SetCVar('blockTrades','0')
		else
			_G['InterfaceOptionsControlsPanelBlockTradesText']:SetFontObject('GameFontHighlightLeft')
			_G['InterfaceOptionsControlsPanelBlockTradesText']:SetText(BLOCK_TRADES)
		end
		if vars.enable.guild and not vars.gen.disable then
			_G['InterfaceOptionsControlsPanelBlockGuildInvitesText']:SetFontObject('GameFontDisableLeft')
			_G['InterfaceOptionsControlsPanelBlockGuildInvitesText']:SetText(BLOCK_GUILD_INVITES..' ('..a..')')
			SetAutoDeclineGuildInvites(false)
		else
			_G['InterfaceOptionsControlsPanelBlockGuildInvitesText']:SetFontObject('GameFontHighlightLeft')
			_G['InterfaceOptionsControlsPanelBlockGuildInvitesText']:SetText(BLOCK_GUILD_INVITES)
		end
		if vars.enable.chan and not vars.gen.disable then
			_G['InterfaceOptionsControlsPanelBlockChatChannelInvitesText']:SetFontObject('GameFontDisableLeft')
			_G['InterfaceOptionsControlsPanelBlockChatChannelInvitesText']:SetText(BLOCK_CHAT_CHANNEL_INVITE..' ('..a..')')
			SetCVar('blockChannelInvites','0')
		else
			_G['InterfaceOptionsControlsPanelBlockChatChannelInvitesText']:SetFontObject('GameFontHighlightLeft')
			_G['InterfaceOptionsControlsPanelBlockChatChannelInvitesText']:SetText(BLOCK_CHAT_CHANNEL_INVITE)
		end
	end
end

-- WoD's options are in the Controls panel, Legion's options are in the Social panel

function blizzardoptionslegion()
	if vars.enable.trade and not vars.gen.disable then
		_G['InterfaceOptionsSocialPanelBlockTradesText']:SetFontObject('GameFontDisableLeft')
		_G['InterfaceOptionsSocialPanelBlockTradesText']:SetText(BLOCK_TRADES..' ('..a..')')
		SetCVar('blockTrades','0')
	else
		_G['InterfaceOptionsSocialPanelBlockTradesText']:SetFontObject('GameFontHighlightLeft')
		_G['InterfaceOptionsSocialPanelBlockTradesText']:SetText(BLOCK_TRADES)
	end
	if vars.enable.guild and not vars.gen.disable then
		_G['InterfaceOptionsSocialPanelBlockGuildInvitesText']:SetFontObject('GameFontDisableLeft')
		_G['InterfaceOptionsSocialPanelBlockGuildInvitesText']:SetText(BLOCK_GUILD_INVITES..' ('..a..')')
		SetAutoDeclineGuildInvites(false)
	else
		_G['InterfaceOptionsSocialPanelBlockGuildInvitesText']:SetFontObject('GameFontHighlightLeft')
		_G['InterfaceOptionsSocialPanelBlockGuildInvitesText']:SetText(BLOCK_GUILD_INVITES)
	end
	if vars.enable.chan and not vars.gen.disable then
		_G['InterfaceOptionsSocialPanelBlockChatChannelInvitesText']:SetFontObject('GameFontDisableLeft')
		_G['InterfaceOptionsSocialPanelBlockChatChannelInvitesText']:SetText(BLOCK_CHAT_CHANNEL_INVITE..' ('..a..')')
		SetCVar('blockChannelInvites','0')
	else
		_G['InterfaceOptionsSocialPanelBlockChatChannelInvitesText']:SetFontObject('GameFontHighlightLeft')
		_G['InterfaceOptionsSocialPanelBlockChatChannelInvitesText']:SetText(BLOCK_CHAT_CHANNEL_INVITE)
	end
end

-- event registration

function register()
	if vars.gen.disable then
		DCL:UnregisterAllEvents()
		StaticPopupDialogs['CHAT_CHANNEL_INVITE']=tcopy(SPD_['CHAT_CHANNEL_INVITE'])
		StaticPopupDialogs['DUEL_REQUESTED']=tcopy(SPD_['DUEL_REQUESTED'])
		StaticPopupDialogs['PET_BATTLE_PVP_DUEL_REQUESTED']=tcopy(SPD_['PET_BATTLE_PVP_DUEL_REQUESTED'])
		StaticPopupDialogs['PARTY_INVITE']=tcopy(SPD_['PARTY_INVITE'])
		StaticPopupDialogs['PARTY_INVITE_XREALM']=tcopy(SPD_['PARTY_INVITE_XREALM'])
		SetLFGFrame('default')
		SetPetitionFrame('default')
		SetTradeFrame('default')
		for i=1,STATICPOPUP_NUMDIALOGS do
			SetStaticPopup(_G['StaticPopup'..i],'default')
		end
		for k1,v1 in next,defs do for k2 in next,v1 do
			if k1 ~= 'chan' and k1 ~= 'open' and k1 ~= 'log' and k2 ~= 'disable' then
			object[k1..k2]:Disable() end
		end end
		for i=1,21 do
			object['text'..format('%02i',i)]:SetFontObject('GameFontDisableSmall')
		end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontDisableSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontNormalLarge')
		object['header01']:SetFontObject('GameFontDisableLarge')
	else
		DCL:RegisterEvent('ADDON_LOADED')
		DCL:RegisterEvent('PLAYER_LOGIN')
		DCL:RegisterEvent('PLAYER_LOGOUT')
		DCL:RegisterEvent('CHAT_MSG_ADDON')
		DCL:RegisterEvent('CHAT_MSG_WHISPER_INFORM')
		DCL:RegisterEvent('CVAR_UPDATE')
		DCL:RegisterEvent('DISABLE_DECLINE_GUILD_INVITE')
		DCL:RegisterEvent('ENABLE_DECLINE_GUILD_INVITE')
		if vars.enable.chan then
			DCL:RegisterEvent('CHANNEL_INVITE_REQUEST')
			StaticPopupDialogs['CHAT_CHANNEL_INVITE']=tcopy(DeclinerDialogs['CHAT_CHANNEL_INVITE'])
		else
			DCL:UnregisterEvent('CHANNEL_INVITE_REQUEST')
			StaticPopupDialogs['CHAT_CHANNEL_INVITE']=tcopy(SPD_['CHAT_CHANNEL_INVITE'])
		end
		if vars.enable.duel then
			DCL:RegisterEvent('DUEL_REQUESTED')
			StaticPopupDialogs['DUEL_REQUESTED']=tcopy(DeclinerDialogs['DUEL_REQUESTED'])
		else
			DCL:UnregisterEvent('DUEL_REQUESTED')
			StaticPopupDialogs['DUEL_REQUESTED']=tcopy(SPD_['DUEL_REQUESTED'])
		end
		if vars.enable.pet then
			DCL:RegisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
			StaticPopupDialogs['PET_BATTLE_PVP_DUEL_REQUESTED']=tcopy(DeclinerDialogs['PET_BATTLE_PVP_DUEL_REQUESTED'])
		else
			DCL:UnregisterEvent('PET_BATTLE_PVP_DUEL_REQUESTED')
			StaticPopupDialogs['PET_BATTLE_PVP_DUEL_REQUESTED']=tcopy(SPD_['PET_BATTLE_PVP_DUEL_REQUESTED'])
		end
		if vars.enable.group then
			DCL:RegisterEvent('PARTY_INVITE_REQUEST')
			StaticPopupDialogs['PARTY_INVITE']=tcopy(DeclinerDialogs['PARTY_INVITE'])
			StaticPopupDialogs['PARTY_INVITE_XREALM']=tcopy(DeclinerDialogs['PARTY_INVITE_XREALM'])
			SetLFGFrame('decline')
		else
			DCL:UnregisterEvent('PARTY_INVITE_REQUEST')
			StaticPopupDialogs['PARTY_INVITE']=tcopy(SPD_['PARTY_INVITE'])
			StaticPopupDialogs['PARTY_INVITE_XREALM']=tcopy(SPD_['PARTY_INVITE_XREALM'])
			SetLFGFrame('default')
		end
		if vars.enable.guild then
			DCL:RegisterEvent('GUILD_INVITE_REQUEST')
		else
			DCL:UnregisterEvent('GUILD_INVITE_REQUEST')
		end
		if vars.enable.chart then
			DCL:RegisterEvent('PETITION_SHOW')
			SetPetitionFrame('decline')
		else
			DCL:UnregisterEvent('PETITION_SHOW')
			SetPetitionFrame('default')
		end
		if vars.enable.trade then
			DCL:RegisterEvent('CHAT_MSG_SYSTEM')
			DCL:RegisterEvent('TRADE_SHOW')
			DCL:RegisterEvent('TRADE_CLOSED')
			SetTradeFrame('decline')
		else
			DCL:UnregisterEvent('CHAT_MSG_SYSTEM')
			DCL:UnregisterEvent('TRADE_SHOW')
			DCL:UnregisterEvent('TRADE_CLOSED')
			SetTradeFrame('default')
		end
		if vars.enable.duel or vars.enable.pet or vars.enable.trade then
			DCL:RegisterEvent('UI_INFO_MESSAGE')
		else
			DCL:UnregisterEvent('UI_INFO_MESSAGE')
		end
		if vars.enable.duel or vars.enable.pet or vars.enable.group or vars.enable.chan then
			for i=1,STATICPOPUP_NUMDIALOGS do
				SetStaticPopup(_G['StaticPopup'..i],'decline')
			end
		else
			for i=1,STATICPOPUP_NUMDIALOGS do
				SetStaticPopup(_G['StaticPopup'..i],'default')
			end
		end
		for k1,v1 in next,defs do for k2 in next,v1 do
			if k1 ~= 'chan' and k1 ~= 'open' and k1 ~= 'log' then
			object[k1..k2]:Enable() end
		end end
		for i=1,21 do
			object['text'..format('%02i',i)]:SetFontObject('GameFontHighlightSmall')
		end
		_G[a..'Button50'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button51'..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[a..'Button52'..'Text']:SetFontObject('GameFontHighlightSmall')
		object['header01']:SetFontObject('GameFontNormalLarge')
	end
	if vars.gen.disable or vars.gen.msgs then
		object['text21']:SetFontObject('GameFontDisableSmall')
		for k1,v1 in next,defs do for k2 in next,v1 do
			if k1 == 'msg' then
			object[k1..k2]:Disable() end
		end end
	else
		object['text21']:SetFontObject('GameFontHighlightSmall')
		for k1,v1 in next,defs do for k2 in next,v1 do
			if k1 == 'msg' then
			object[k1..k2]:Enable() end
		end end
	end
	if vars.gen.disable or vars.gen.openinv then
		openinvites=false
		object['OIinput']:Disable()
		object['OIlist']:Disable()
		object['OIbutton']:Disable()
		DCL:UnregisterEvent('CHAT_MSG_CHANNEL')
		DCL:UnregisterEvent('CHAT_MSG_EMOTE')
		DCL:UnregisterEvent('CHAT_MSG_SAY')
		DCL:UnregisterEvent('CHAT_MSG_YELL')
		DCL:UnregisterEvent('GROUP_ROSTER_UPDATE')
		DCL:UnregisterEvent('ZONE_CHANGED_NEW_AREA')
		UIDropDownMenu_DisableDropDown(object['OIdrop'])
		object['OIinput']:SetFontObject('GameFontDisableSmall')
		object['OIlist']:SetFontObject('GameFontDisableSmall')
	else
		object['OIinput']:Enable()
		object['OIlist']:Enable()
		object['OIbutton']:Enable()
		DCL:RegisterEvent('CHAT_MSG_CHANNEL')
		DCL:RegisterEvent('CHAT_MSG_EMOTE')
		DCL:RegisterEvent('CHAT_MSG_SAY')
		DCL:RegisterEvent('CHAT_MSG_YELL')
		DCL:RegisterEvent('GROUP_ROSTER_UPDATE')
		DCL:RegisterEvent('ZONE_CHANGED_NEW_AREA')
		UIDropDownMenu_EnableDropDown(object['OIdrop'])
		object['OIinput']:SetFontObject('GameFontHighlightSmall')
		object['OIlist']:SetFontObject('GameFontHighlightSmall')
	end
end

-- Interface Options panel
-- I would like to thank FlexYourHead of the addon Work_Complete for a ton of comments in their code
-- It made understanding buttons, text, and the options frame a whole lot easier

function create(obj,xloc,yloc,misc,text,image,rinset)
	if obj == 'b' then
		numbutton=numbutton+1
		local name=a..'Button'..format('%02i',numbutton)
		obj=CreateFrame('CheckButton',name,DCLPanel,'OptionsCheckButtonTemplate')
		obj:SetPoint('CENTER',DCLPanel,'TOPRIGHT',xloc,yloc)
		obj:SetHitRectInsets(-10,rinset or -10,-10,-10)
		obj:SetCheckedTexture(file[image])
		obj:SetDisabledCheckedTexture(file[image])
		obj:GetCheckedTexture():SetVertexColor(0.8,0.8,0.8)
		local supported=obj:GetDisabledCheckedTexture():SetDesaturated(true)
		if not supported then obj:GetDisabledCheckedTexture():SetAlpha(0.4) end
		obj:GetDisabledCheckedTexture():SetVertexColor(0.4,0.4,0.4)
		obj.tooltipText=misc
		_G[name..'Text']:SetFontObject('GameFontHighlightSmall')
		_G[name..'Text']:SetText(text)
	elseif obj == 't' then
		numtext=numtext+1
		local name=a..'Text'..format('%02i',numtext)
		obj=DCLPanel:CreateFontString(name,'ARTWORK','GameFontHighlightSmall')
		obj:SetPoint(misc,DCLPanel,'TOPRIGHT',xloc,yloc)
		obj:SetText(text)
	end
	return obj
end

function openinvlist()
	sort(vars.open)
	local text
	for i=1,#vars.open do
		if not text then text=vars.open[i]
		else text=text..'\n'..vars.open[i] end
	end
	object['OIlist']:SetText(text or '')
end

function setoptions()
	for k1,v1 in next,defs do for k2 in next,v1 do
		if k1 ~= 'chan' and k1 ~= 'open' and k1 ~= 'log' then
		object[k1..k2]:SetChecked(vars[k1][k2]) end
	end end
end

function options()
	DCLPanel.name=a
	InterfaceOptions_AddCategory(DCLPanel,a)
	object['header01']=DCLPanel:CreateFontString(a..'Header','ARTWORK','GameFontNormalLarge')
	object['header01']:SetPoint('TOPLEFT',16,-16)
	object['header01']:SetText(a)
	object['line01']=DCLPanel:CreateTexture(a..'Separator1')
	object['line01']:SetSize(600,1)
	object['line01']:SetPoint('LEFT',DCLPanel,'TOPLEFT',10,-40)
	if tostring(format('%4$s',GetBuildInfo())) ~= '60200' then
	object['line01']:SetColorTexture(0.25,0.25,0.25) else
	object['line01']:SetTexture(0.25,0.25,0.25) end -- WoD uses SetTexture, Legion uses SetColorTexture
	object['line02']=DCLPanel:CreateTexture(a..'Separator2')
	object['line02']:SetSize(600,1)
	object['line02']:SetPoint('LEFT',DCLPanel,'TOPLEFT',10,-428)
	if tostring(format('%4$s',GetBuildInfo())) ~= '60200' then
	object['line02']:SetColorTexture(0.25,0.25,0.25) else
	object['line02']:SetTexture(0.25,0.25,0.25) end -- WoD uses SetTexture, Legion uses SetColorTexture
	object['text01']=create('t',-460,-64,'CENTER',S['DUE'])
	object['text02']=create('t',-460,-80,'CENTER',S['REQ'])
	object['text03']=create('t',-392,-64,'CENTER',S['PBA'])
	object['text04']=create('t',-392,-80,'CENTER',S['REQ'])
	object['text05']=create('t',-324,-64,'CENTER',S['GLD'])
	object['text06']=create('t',-324,-80,'CENTER',S['INV'])
	object['text07']=create('t',-256,-64,'CENTER',S['GRO'])
	object['text08']=create('t',-256,-80,'CENTER',S['INV'])
	object['text09']=create('t',-188,-64,'CENTER',S['GLD'])
	object['text10']=create('t',-188,-80,'CENTER',S['PTN'])
	object['text11']=create('t',-120,-64,'CENTER',S['TRA'])
	object['text12']=create('t',-120,-80,'CENTER',S['REQ'])
	object['text13']=create('t',-52,-64,'CENTER',S['CHA'])
	object['text14']=create('t',-52,-80,'CENTER',S['INV'])
	object['text15']=create('t',-492,-112,'RIGHT',S['EDE'])
	object['text16']=create('t',-492,-160,'RIGHT',S['WHI'])
	object['text17']=create('t',-492,-208,'RIGHT',S['GRM'])
	object['text18']=create('t',-492,-256,'RIGHT',S['GME'])
	object['text19']=create('t',-492,-304,'RIGHT',S['RFR'])
	object['text20']=create('t',-492,-352,'RIGHT',S['BNF'])
	object['text21']=create('t',-492,-400,'RIGHT',S['PLM'])
	object['enableduel']=create('b',-460,-112,g..L['ENABLE_DUEL'],'','ch')
	object['enablepet']=create('b',-392,-112,g..L['ENABLE_PET'],'','ch')
	object['enableguild']=create('b',-324,-112,g..L['ENABLE_GUILD'],'','ch')
	object['enablegroup']=create('b',-256,-112,g..L['ENABLE_GROUP'],'','ch')
	object['enablechart']=create('b',-188,-112,g..L['ENABLE_CHART'],'','ch')
	object['enabletrade']=create('b',-120,-112,g..L['ENABLE_TRADE'],'','ch')
	object['enablechan']=create('b',-52,-112,g..L['ENABLE_CHAN'],'','ch')
	object['tellduel']=create('b',-460,-160,r..L['TELL_DUEL'],'','cr')
	object['tellpet']=create('b',-392,-160,r..L['TELL_PET'],'','cr')
	object['tellguild']=create('b',-324,-160,r..L['TELL_GUILD'],'','cr')
	object['tellgroup']=create('b',-256,-160,r..L['TELL_GROUP'],'','cr')
	object['tellchart']=create('b',-188,-160,r..L['TELL_CHART'],'','cr')
	object['telltrade']=create('b',-120,-160,r..L['TELL_TRADE'],'','cr')
	object['tellchan']=create('b',-52,-160,r..L['TELL_CHAN'],'','cr')
	object['groupduel']=create('b',-460,-208,r..L['GROUP_DUEL'],'','cr')
	object['grouppet']=create('b',-392,-208,r..L['GROUP_PET'],'','cr')
	object['groupguild']=create('b',-324,-208,r..L['GROUP_GUILD'],'','cr')
	object['groupgroup']=create('b',-256,-208,r..L['GROUP_GROUP'],'','cr')
	object['groupchart']=create('b',-188,-208,r..L['GROUP_CHART'],'','cr')
	object['grouptrade']=create('b',-120,-208,r..L['GROUP_TRADE'],'','cr')
	object['groupchan']=create('b',-52,-208,r..L['GROUP_CHAN'],'','cr')
	object['guildduel']=create('b',-460,-256,r..L['GUILD_DUEL'],'','cr')
	object['guildpet']=create('b',-392,-256,r..L['GUILD_PET'],'','cr')
	object['guildguild']=create('b',-324,-256,r..L['GUILD_GUILD'],'','cr')
	object['guildgroup']=create('b',-256,-256,r..L['GUILD_GROUP'],'','cr')
	object['guildchart']=create('b',-188,-256,r..L['GUILD_CHART'],'','cr')
	object['guildtrade']=create('b',-120,-256,r..L['GUILD_TRADE'],'','cr')
	object['guildchan']=create('b',-52,-256,r..L['GUILD_CHAN'],'','cr')
	object['friendduel']=create('b',-460,-304,r..L['FRIEND_DUEL'],'','cr')
	object['friendpet']=create('b',-392,-304,r..L['FRIEND_PET'],'','cr')
	object['friendguild']=create('b',-324,-304,r..L['FRIEND_GUILD'],'','cr')
	object['friendgroup']=create('b',-256,-304,r..L['FRIEND_GROUP'],'','cr')
	object['friendchart']=create('b',-188,-304,r..L['FRIEND_CHART'],'','cr')
	object['friendtrade']=create('b',-120,-304,r..L['FRIEND_TRADE'],'','cr')
	object['friendchan']=create('b',-52,-304,r..L['FRIEND_CHAN'],'','cr')
	object['bnetduel']=create('b',-460,-352,r..L['BNET_DUEL'],'','cr')
	object['bnetpet']=create('b',-392,-352,r..L['BNET_PET'],'','cr')
	object['bnetguild']=create('b',-324,-352,r..L['BNET_GUILD'],'','cr')
	object['bnetgroup']=create('b',-256,-352,r..L['BNET_GROUP'],'','cr')
	object['bnetchart']=create('b',-188,-352,r..L['BNET_CHART'],'','cr')
	object['bnettrade']=create('b',-120,-352,r..L['BNET_TRADE'],'','cr')
	object['bnetchan']=create('b',-52,-352,r..L['BNET_CHAN'],'','cr')
	object['msgduel']=create('b',-460,-400,r..L['MSG_DUEL'],'','cr')
	object['msgpet']=create('b',-392,-400,r..L['MSG_PET'],'','cr')
	object['msgguild']=create('b',-324,-400,r..L['MSG_GUILD'],'','cr')
	object['msggroup']=create('b',-256,-400,r..L['MSG_GROUP'],'','cr')
	object['msgchart']=create('b',-188,-400,r..L['MSG_CHART'],'','cr')
	object['msgtrade']=create('b',-120,-400,r..L['MSG_TRADE'],'','cr')
	object['msgchan']=create('b',-52,-400,r..L['MSG_CHAN'],'','cr')
	object['genopeninv']=create('b',-256,-456,r..L['X_DISABLE_OPENINV'],S['DOI']..S['INV'],'cr',-200)
	object['genmsgs']=create('b',-256,-496,r..L['X_DISABLE_MESSAGES'],S['DDM'],'cr',-200)
	object['gendisable']=create('b',-256,-536,r..L['X_DISABLE_DECLINER'],S['DDA'],'cr',-200)

	for k1,v1 in next,defs do for k2 in next,v1 do
		if k1 == k2 then
			object[k1..k2]:SetScript('PostClick',function(self)
				if gsnd[k1] then StopSound(gsnd[k1]) end
				local x=random(1,55555)
				_,gsnd[k1]=PlaySoundKitID(x)
				self:SetChecked(false)
			end)
		elseif k1 ~= 'chan' and k1 ~= 'open' and k1 ~= 'log' then
			object[k1..k2]:SetScript('PostClick',function(self)
				vars[k1][k2]=self:GetChecked()
				blizzardoptions()
				register()
			end)
		end
	end end

-- the following inputbox, button, and scrolling list inspired by funkydude's BadBoy_CCleaner chat filter options

	object['OIinput']=CreateFrame('EditBox',a..'OpenInvInput',DCLPanel,'InputBoxTemplate')
	object['OIinput']:SetPoint('TOPRIGHT',-300,-461)
	object['OIinput']:SetAutoFocus(false)
	object['OIinput']:EnableMouse(true)
	object['OIinput']:SetHeight(20)
	object['OIinput']:SetWidth(112)
	object['OIinput']:SetScript('OnEnterPressed',function()object['OIbutton']:Click()end)
	object['OIinput']:SetScript('OnEscapePressed',function(frame)
		frame:SetText('')
		frame:ClearFocus()
	end)
	object['OIinput']:SetScript('OnEnter',function(self)
		if vars.gen.disable or vars.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L['X_OPENINV_INPUT'],nil,nil,nil,nil,true)
		end
	end)
	object['OIinput']:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	object['OIbutton']=CreateFrame('Button',a..'OpenInvButton',DCLPanel,'UIPanelButtonTemplate')
	object['OIbutton']:SetPoint('TOPRIGHT',object['OIinput'],'BOTTOMRIGHT',1,0)
	object['OIbutton']:SetPoint('TOPLEFT',object['OIinput'],'BOTTOMLEFT',-6,0)
	object['OIbutton']:SetNormalFontObject('GameFontNormalSmall')
	object['OIbutton']:SetHighlightFontObject('GameFontHighlightSmall')
	object['OIbutton']:SetDisabledFontObject('GameFontDisableSmall')
	object['OIbutton']:SetText(S['ADR'])
	object['OIbutton']:SetScript('OnClick',function()
		object['OIinput']:ClearFocus()
		local t=object['OIinput']:GetText()
		if t == '' or t:find('^ *$') then object['OIinput']:SetText('') return end
		local found
		for i=1,#vars.open do if vars.open[i] == t then found=i end end
		if found then tremove(vars.open,found)
		else tinsert(vars.open,t) end
		openinvlist()
		object['OIinput']:SetText('')
	end)
	object['OIbutton']:SetScript('OnEnter',function(self)
		if vars.gen.disable or vars.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-1)
			GameTooltip:SetText(L['X_OPENINV_BUTTON'],nil,nil,nil,nil,true)
		end
	end)
	object['OIbutton']:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	object['OIdrop']=CreateFrame('Frame',a..'OpenInvDropDown',DCLPanel,'UIDropDownMenuTemplate')
	object['OIdrop']:SetPoint('TOPLEFT',object['OIbutton'],'BOTTOMLEFT',-16,0)
	UIDropDownMenu_SetWidth(object['OIdrop'],100)
	UIDropDownMenu_SetText(object['OIdrop'],S['CHS'])
	UIDropDownMenu_Initialize(object['OIdrop'],function(frame)
		local info=UIDropDownMenu_CreateInfo()
		info.func=function(self,arg1)vars.chan[arg1]=not vars.chan[arg1]end
		info.text,info.checked,info.arg1=S['EMO'],vars.chan.emote,'emote'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=S['SAY'],vars.chan.say,'say'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=S['YEL'],vars.chan.yell,'yell'
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=S['GEN'],vars.chan[1],1
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=S['TRA'],vars.chan[2],2
		UIDropDownMenu_AddButton(info)
		info.text,info.checked,info.arg1=S['LFG'],vars.chan[26],26
		UIDropDownMenu_AddButton(info)
	end)
	object['OIdrop']:SetScript('OnEnter',function(self)
		if vars.gen.disable or vars.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT',-16)
			GameTooltip:SetText(L['X_OPENINV_MENU'],nil,nil,nil,nil,true)
		end
	end)
	object['OIdrop']:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	object['OIback']=CreateFrame('Frame',a..'OpenInvBackdrop',DCLPanel)
	object['OIback']:SetPoint('TOPRIGHT',object['OIinput'],'TOPLEFT',-24,0)
	object['OIback']:SetPoint('BOTTOMLEFT',object['OIdrop'],'BOTTOMLEFT',-119,6)
	object['OIback']:SetBackdrop({
		bgFile='Interface\\Tooltips\\UI-Tooltip-Background',
		edgeFile='Interface\\Tooltips\\UI-Tooltip-Border',
		tile=true,tileSize=8,edgeSize=8,
		insets={left=2,right=2,top=2,bottom=2}})
	object['OIback']:SetBackdropColor(0,0,0,.75)
	object['OIback']:SetBackdropBorderColor(.25,.25,.25)
	object['OIback']:SetScript('OnEnter',function(self)
		if vars.gen.disable or vars.gen.openinv then
			GameTooltip:Hide()
		else
			GameTooltip:SetOwner(self,self.tooltipOwnerPoint or 'ANCHOR_RIGHT')
			GameTooltip:SetText(L['X_OPENINV_LIST'],nil,nil,nil,nil,true)
		end
	end)
	object['OIback']:SetScript('OnLeave',function(self)
		GameTooltip:Hide()
	end)

	object['OIscroll']=CreateFrame('ScrollFrame',a..'OpenInvScroll',DCLPanel,'UIPanelScrollFrameTemplate')
	object['OIscroll']:SetPoint('LEFT',object['OIback'],'LEFT',4.5,0)
	object['OIscroll']:SetPoint('RIGHT',object['OIback'],'RIGHT',-4.5,0)
	object['OIscroll']:SetPoint('TOP',object['OIback'],'TOP',0,-4)
	object['OIscroll']:SetPoint('BOTTOM',object['OIback'],'BOTTOM',0,4)

	object['OIlist']=CreateFrame('EditBox',a..'OpenInvList',DCLPanel)
	object['OIlist']:SetMaxLetters(1000)
	object['OIlist']:SetAutoFocus(false)
	object['OIlist']:EnableMouse(false)
	object['OIlist']:SetMultiLine(true)
	object['OIlist']:SetWidth(120)
	object['OIlist']:Show()

	object['OIscroll']:SetScrollChild(object['OIlist'])

	DCLPanel.okay=function()
		Decliner_SV=nil
		Decliner_SV=tcopy(vars)
		notemp=true
	end
	DCLPanel.cancel=function()
		vars=tcopy(temp)
		register()
		notemp=true
	end
	DCLPanel.default=function()
		local tmp=vars.log
		vars=nil
		vars=tcopy(defs)
		vars.log=tcopy(tmp)
		Decliner_SV=nil
		Decliner_SV=tcopy(vars)
		openinvlist()
		setoptions()
		register()
		blizzardoptions()
	end
	DCLPanel.refresh=function()
		openinvlist()
		setoptions()
		blizzardoptions()
		if notemp then temp=tcopy(vars) notemp=false end
	end
	SLASH_DECLINER_1='/decliner'
	SLASH_DECLINER_2='/decline'
	SLASH_DECLINER_3='/dcl'
	SlashCmdList['DECLINER_']=function()
	-- thank you to https://www.townlong-yak.com/framexml/bugs/6 for this extremely simple workaround
		InterfaceOptionsFrame_OpenToCategory(a)
		InterfaceOptionsFrame_OpenToCategory(a)
	end
end